@@ -12,13 +12,18 @@
 // limitations under the License.
 
 //go:build !skip_store_tests
-// +build !skip_store_tests
 
 package server
 
 import (
 	"fmt"
+	"os"
+	"path/filepath"
+	"reflect"
 	"testing"
+	"time"
+
+	"github.com/nats-io/nats-server/v2/server/gsl"
 )
 
 func testAllStoreAllPermutations(t *testing.T, compressionAndEncryption bool, cfg StreamConfig, fn func(t *testing.T, fs StreamStore)) {
@@ -57,13 +62,13 @@ func TestStoreMsgLoadNextMsgMulti(t *testing.T) {
 			// Put 1k msgs in
 			for i := 0; i < 1000; i++ {
 				subj := fmt.Sprintf("foo.%d", i)
-				fs.StoreMsg(subj, nil, []byte("ZZZ"))
+				fs.StoreMsg(subj, nil, []byte("ZZZ"), 0)
 			}
 
 			var smv StoreMsg
 			// Do multi load next with 1 wc entry.
-			sl := NewSublistWithCache()
-			sl.Insert(&subscription{subject: []byte("foo.>")})
+			sl := gsl.NewSublist[struct{}]()
+			sl.Insert("foo.>", struct{}{})
 			for i, seq := 0, uint64(1); i < 1000; i++ {
 				sm, nseq, err := fs.LoadNextMsgMulti(sl, seq, &smv)
 				require_NoError(t, err)
@@ -73,10 +78,10 @@ func TestStoreMsgLoadNextMsgMulti(t *testing.T) {
 			}
 
 			// Now do multi load next with 1000 literal subjects.
-			sl = NewSublistWithCache()
+			sl = gsl.NewSublist[struct{}]()
 			for i := 0; i < 1000; i++ {
 				subj := fmt.Sprintf("foo.%d", i)
-				sl.Insert(&subscription{subject: []byte(subj)})
+				sl.Insert(subj, struct{}{})
 			}
 			for i, seq := 0, uint64(1); i < 1000; i++ {
 				sm, nseq, err := fs.LoadNextMsgMulti(sl, seq, &smv)
@@ -87,10 +92,10 @@ func TestStoreMsgLoadNextMsgMulti(t *testing.T) {
 			}
 
 			// Check that we can pull out 3 individuals.
-			sl = NewSublistWithCache()
-			sl.Insert(&subscription{subject: []byte("foo.2")})
-			sl.Insert(&subscription{subject: []byte("foo.222")})
-			sl.Insert(&subscription{subject: []byte("foo.999")})
+			sl = gsl.NewSublist[struct{}]()
+			sl.Insert("foo.2", struct{}{})
+			sl.Insert("foo.222", struct{}{})
+			sl.Insert("foo.999", struct{}{})
 			sm, seq, err := fs.LoadNextMsgMulti(sl, 1, &smv)
 			require_NoError(t, err)
 			require_Equal(t, sm.subj, "foo.2")
@@ -168,7 +173,7 @@ func TestStoreSubjectStateConsistency(t *testing.T) {
 
 			// Publish an initial batch of messages.
 			for i := 0; i < 4; i++ {
-				_, _, err := fs.StoreMsg("foo", nil, nil)
+				_, _, err := fs.StoreMsg("foo", nil, nil, 0)
 				require_NoError(t, err)
 			}
 
@@ -221,7 +226,7 @@ func TestStoreSubjectStateConsistency(t *testing.T) {
 
 			// Publish some more messages so we can test another scenario.
 			for i := 0; i < 3; i++ {
-				_, _, err := fs.StoreMsg("foo", nil, nil)
+				_, _, err := fs.StoreMsg("foo", nil, nil, 0)
 				require_NoError(t, err)
 			}
 
@@ -257,13 +262,13 @@ func TestStoreSubjectStateConsistency(t *testing.T) {
 			expectLastSeq(6)
 
 			// We store a new message for ss.Last and remove it after, which marks it to be recalculated.
-			_, _, err = fs.StoreMsg("foo", nil, nil)
+			_, _, err = fs.StoreMsg("foo", nil, nil, 0)
 			require_NoError(t, err)
 			removed, err = fs.RemoveMsg(8)
 			require_NoError(t, err)
 			require_True(t, removed)
 			// This will be the new ss.Last message, so reset ss.lastNeedsUpdate
-			_, _, err = fs.StoreMsg("foo", nil, nil)
+			_, _, err = fs.StoreMsg("foo", nil, nil, 0)
 			require_NoError(t, err)
 
 			// ss.First should remain the same, but ss.Last should equal the last message.
@@ -285,7 +290,7 @@ func TestStoreSubjectStateConsistencyOptimization(t *testing.T) {
 			fillMsgs := func(c int) {
 				t.Helper()
 				for i := 0; i < c; i++ {
-					_, _, err := fs.StoreMsg("foo", nil, nil)
+					_, _, err := fs.StoreMsg("foo", nil, nil, 0)
 					require_NoError(t, err)
 				}
 			}
@@ -376,7 +381,7 @@ func TestStoreMaxMsgsPerUpdateBug(t *testing.T) {
 		t, false, config(),
 		func(t *testing.T, fs StreamStore) {
 			for i := 0; i < 5; i++ {
-				_, _, err := fs.StoreMsg("foo", nil, nil)
+				_, _, err := fs.StoreMsg("foo", nil, nil, 0)
 				require_NoError(t, err)
 			}
 
@@ -435,7 +440,7 @@ func TestStoreCompactCleansUpDmap(t *testing.T) {
 
 					// Publish messages, should have no interior deletes.
 					for i := 0; i < 3; i++ {
-						_, _, err := fs.StoreMsg("foo", nil, nil)
+						_, _, err := fs.StoreMsg("foo", nil, nil, 0)
 						require_NoError(t, err)
 					}
 					require_Len(t, dmapEntries(), 0)
@@ -486,7 +491,7 @@ func TestStoreTruncateCleansUpDmap(t *testing.T) {
 
 					// Publish messages, should have no interior deletes.
 					for i := 0; i < 3; i++ {
-						_, _, err := fs.StoreMsg("foo", nil, nil)
+						_, _, err := fs.StoreMsg("foo", nil, nil, 0)
 						require_NoError(t, err)
 					}
 					require_Len(t, dmapEntries(), 0)
@@ -513,3 +518,287 @@ func TestStoreTruncateCleansUpDmap(t *testing.T) {
 		})
 	}
 }
+
+// https://github.com/nats-io/nats-server/issues/6709
+func TestStorePurgeExZero(t *testing.T) {
+	testAllStoreAllPermutations(
+		t, false,
+		StreamConfig{Name: "TEST", Subjects: []string{"foo"}},
+		func(t *testing.T, fs StreamStore) {
+			// Simple purge all.
+			_, err := fs.Purge()
+			require_NoError(t, err)
+			ss := fs.State()
+			require_Equal(t, ss.FirstSeq, 1)
+			require_Equal(t, ss.LastSeq, 0)
+
+			// PurgeEx(seq=0) must be equal.
+			_, err = fs.PurgeEx(_EMPTY_, 0, 0)
+			require_NoError(t, err)
+			ss = fs.State()
+			require_Equal(t, ss.FirstSeq, 1)
+			require_Equal(t, ss.LastSeq, 0)
+		},
+	)
+}
+
+func TestStoreUpdateConfigTTLState(t *testing.T) {
+	config := func() StreamConfig {
+		return StreamConfig{Name: "TEST", Subjects: []string{"foo"}}
+	}
+	testAllStoreAllPermutations(
+		t, false, config(),
+		func(t *testing.T, fs StreamStore) {
+			cfg := config()
+			switch fs.(type) {
+			case *fileStore:
+				cfg.Storage = FileStorage
+			case *memStore:
+				cfg.Storage = MemoryStorage
+			}
+
+			// TTLs disabled at this point so this message should survive.
+			seq, _, err := fs.StoreMsg("foo", nil, nil, 1)
+			require_NoError(t, err)
+			time.Sleep(2 * time.Second)
+			_, err = fs.LoadMsg(seq, nil)
+			require_NoError(t, err)
+
+			// Now enable TTLs.
+			cfg.AllowMsgTTL = true
+			require_NoError(t, fs.UpdateConfig(&cfg))
+
+			// TTLs enabled at this point so this message should be cleaned up.
+			seq, _, err = fs.StoreMsg("foo", nil, nil, 1)
+			require_NoError(t, err)
+			time.Sleep(2 * time.Second)
+			_, err = fs.LoadMsg(seq, nil)
+			require_Error(t, err)
+
+			// Now disable TTLs again.
+			cfg.AllowMsgTTL = false
+			require_NoError(t, fs.UpdateConfig(&cfg))
+
+			// TTLs disabled again so this message should survive.
+			seq, _, err = fs.StoreMsg("foo", nil, nil, 1)
+			require_NoError(t, err)
+			time.Sleep(2 * time.Second)
+			_, err = fs.LoadMsg(seq, nil)
+			require_NoError(t, err)
+		},
+	)
+}
+
+func TestStoreStreamInteriorDeleteAccounting(t *testing.T) {
+	tests := []struct {
+		title  string
+		action func(s StreamStore, lseq uint64)
+	}{
+		{
+			title: "TruncateWithRemove",
+			action: func(s StreamStore, lseq uint64) {
+				seq, _, err := s.StoreMsg("foo", nil, nil, 0)
+				require_NoError(t, err)
+				require_Equal(t, seq, lseq)
+				removed, err := s.RemoveMsg(lseq)
+				require_NoError(t, err)
+				require_True(t, removed)
+				require_NoError(t, s.Truncate(lseq))
+			},
+		},
+		{
+			title: "TruncateWithErase",
+			action: func(s StreamStore, lseq uint64) {
+				seq, _, err := s.StoreMsg("foo", nil, nil, 0)
+				require_NoError(t, err)
+				require_Equal(t, seq, lseq)
+				removed, err := s.EraseMsg(lseq)
+				require_NoError(t, err)
+				require_True(t, removed)
+				require_NoError(t, s.Truncate(lseq))
+			},
+		},
+		{
+			title: "TruncateWithTombstone",
+			action: func(s StreamStore, lseq uint64) {
+				seq, _, err := s.StoreMsg("foo", nil, nil, 0)
+				require_NoError(t, err)
+				require_Equal(t, seq, lseq)
+				if fs, ok := s.(*fileStore); ok {
+					removed, err := fs.removeMsg(lseq, false, false, true)
+					require_NoError(t, err)
+					require_True(t, removed)
+				} else {
+					removed, err := s.RemoveMsg(lseq)
+					require_NoError(t, err)
+					require_True(t, removed)
+				}
+				require_NoError(t, s.Truncate(lseq))
+			},
+		},
+		{
+			title: "SkipMsg",
+			action: func(s StreamStore, lseq uint64) {
+				s.SkipMsg(0)
+			},
+		},
+		{
+			title: "SkipMsgs",
+			action: func(s StreamStore, lseq uint64) {
+				require_NoError(t, s.SkipMsgs(lseq, 1))
+			},
+		},
+	}
+	for _, empty := range []bool{false, true} {
+		for _, test := range tests {
+			t.Run(fmt.Sprintf("Empty=%v/%s", empty, test.title), func(t *testing.T) {
+				cfg := StreamConfig{Name: "zzz", Subjects: []string{"foo"}}
+				testAllStoreAllPermutations(t, true, cfg, func(t *testing.T, s StreamStore) {
+					var err error
+					var lseq uint64
+					if !empty {
+						lseq, _, err = s.StoreMsg("foo", nil, nil, 0)
+						require_NoError(t, err)
+						require_Equal(t, lseq, 1)
+					}
+					lseq++
+
+					test.action(s, lseq)
+
+					// Confirm state as baseline.
+					before := s.State()
+					if empty {
+						require_Equal(t, before.Msgs, 0)
+						require_Equal(t, before.FirstSeq, 2)
+						require_Equal(t, before.LastSeq, 1)
+					} else {
+						require_Equal(t, before.Msgs, 1)
+						require_Equal(t, before.FirstSeq, 1)
+						require_Equal(t, before.LastSeq, 2)
+					}
+
+					var fs *fileStore
+					var ok bool
+					if fs, ok = s.(*fileStore); !ok {
+						return
+					}
+					cfg.Storage = FileStorage
+					fcfg := fs.fcfg
+					created := time.Time{}
+
+					// Restart should equal state.
+					require_NoError(t, fs.Stop())
+					fs, err = newFileStoreWithCreated(fcfg, cfg, created, prf(&fcfg), nil)
+					require_NoError(t, err)
+					defer fs.Stop()
+
+					if state := fs.State(); !reflect.DeepEqual(state, before) {
+						t.Fatalf("Expected state of:\n%+v, got:\n%+v", before, state)
+					}
+
+					// Stop and remove stream state file.
+					require_NoError(t, fs.Stop())
+					require_NoError(t, os.Remove(filepath.Join(fs.fcfg.StoreDir, msgDir, streamStreamStateFile)))
+
+					// Recovering based on blocks should result in the same state.
+					fs, err = newFileStoreWithCreated(fcfg, cfg, created, prf(&fcfg), nil)
+					require_NoError(t, err)
+					defer fs.Stop()
+
+					if state := fs.State(); !reflect.DeepEqual(state, before) {
+						t.Fatalf("Expected state of:\n%+v, got:\n%+v", before, state)
+					}
+
+					// Rebuilding state must also result in the same state.
+					fs.rebuildState(nil)
+					if state := fs.State(); !reflect.DeepEqual(state, before) {
+						t.Fatalf("Expected state of:\n%+v, got:\n%+v", before, state)
+					}
+				})
+			})
+		}
+	}
+}
+
+func TestStoreMsgLoadPrevMsgMulti(t *testing.T) {
+	testAllStoreAllPermutations(
+		t, false,
+		StreamConfig{Name: "zzz", Subjects: []string{"foo.*"}},
+		func(t *testing.T, fs StreamStore) {
+			// Put 1k msgs in
+			for i := range 1000 {
+				subj := fmt.Sprintf("foo.%d", i+1)
+				fs.StoreMsg(subj, nil, []byte("ZZZ"), 0)
+			}
+
+			var sm StoreMsg
+			var count int
+			var state StreamState
+			fs.FastState(&state)
+
+			sl := gsl.NewSimpleSublist()
+			sl.Insert("foo.5", struct{}{})
+			sl.Insert("foo.15", struct{}{})
+			sl.Insert("foo.105", struct{}{})
+
+			for seq := state.LastSeq; seq > 5; seq-- {
+				var err error
+				_, seq, err = fs.LoadPrevMsgMulti(sl, seq, &sm)
+				require_NoError(t, err)
+				require_Equal(t, sm.subj, fmt.Sprintf("foo.%d", sm.seq))
+				count++
+			}
+
+			_, _, err := fs.LoadPrevMsgMulti(sl, 4, &sm)
+			require_Error(t, err, ErrStoreEOF)
+			require_Equal(t, count, 3)
+		},
+	)
+}
+
+func TestStoreGetSeqFromTimeWithInteriorDeletesGap(t *testing.T) {
+	testAllStoreAllPermutations(
+		t, false,
+		StreamConfig{Name: "zzz", Subjects: []string{"foo"}},
+		func(t *testing.T, fs StreamStore) {
+			var start int64
+			for i := range 10 {
+				_, ts, err := fs.StoreMsg("foo", nil, nil, 0)
+				require_NoError(t, err)
+				if i == 1 {
+					start = ts
+				}
+			}
+			// Create a delete gap to prove a simple binary search between sequences
+			// does not work, and deletes need to be accounted for. A simple binary search
+			// will hit the deleted sequences and then return the last sequence.
+			for seq := uint64(4); seq <= 7; seq++ {
+				_, err := fs.RemoveMsg(seq)
+				require_NoError(t, err)
+			}
+			ts := time.Unix(0, start).UTC()
+			require_Equal(t, fs.GetSeqFromTime(ts), 2)
+		},
+	)
+}
+
+func TestStoreGetSeqFromTimeWithTrailingDeletes(t *testing.T) {
+	testAllStoreAllPermutations(
+		t, false,
+		StreamConfig{Name: "zzz", Subjects: []string{"foo"}},
+		func(t *testing.T, fs StreamStore) {
+			var start int64
+			for i := range 3 {
+				_, ts, err := fs.StoreMsg("foo", nil, nil, 0)
+				require_NoError(t, err)
+				if i == 1 {
+					start = ts
+				}
+			}
+			_, err := fs.RemoveMsg(3)
+			require_NoError(t, err)
+			ts := time.Unix(0, start).UTC()
+			require_Equal(t, fs.GetSeqFromTime(ts), 2)
+		},
+	)
+}
