@@ -1,4 +1,4 @@
-// Copyright 2016-2024 The NATS Authors
+// Copyright 2016-2025 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -26,7 +26,6 @@ import (
 	"testing"
 	"time"
 
-	"github.com/nats-io/nats-server/v2/server/stree"
 	"github.com/nats-io/nuid"
 )
 
@@ -685,24 +684,24 @@ func TestSubjectIsLiteral(t *testing.T) {
 }
 
 func TestValidateDestinationSubject(t *testing.T) {
-	checkError(ValidateMappingDestination("foo"), nil, t)
-	checkError(ValidateMappingDestination("foo.bar"), nil, t)
-	checkError(ValidateMappingDestination("*"), nil, t)
-	checkError(ValidateMappingDestination(">"), nil, t)
-	checkError(ValidateMappingDestination("foo.*"), nil, t)
-	checkError(ValidateMappingDestination("foo.>"), nil, t)
-	checkError(ValidateMappingDestination("foo.*.>"), nil, t)
-	checkError(ValidateMappingDestination("foo.*.bar"), nil, t)
-	checkError(ValidateMappingDestination("foo.bar.>"), nil, t)
-	checkError(ValidateMappingDestination("foo.{{wildcard(1)}}"), nil, t)
-	checkError(ValidateMappingDestination("foo.{{ wildcard(1) }}"), nil, t)
-	checkError(ValidateMappingDestination("foo.{{wildcard( 1 )}}"), nil, t)
-	checkError(ValidateMappingDestination("foo.{{partition(2,1)}}"), nil, t)
-	checkError(ValidateMappingDestination("foo.{{SplitFromLeft(2,1)}}"), nil, t)
-	checkError(ValidateMappingDestination("foo.{{SplitFromRight(2,1)}}"), nil, t)
-	checkError(ValidateMappingDestination("foo.{{unknown(1)}}"), ErrInvalidMappingDestination, t)
-	checkError(ValidateMappingDestination("foo..}"), ErrInvalidMappingDestination, t)
-	checkError(ValidateMappingDestination("foo. bar}"), ErrInvalidMappingDestinationSubject, t)
+	checkError(ValidateMapping("bar", "foo"), nil, t)
+	checkError(ValidateMapping("foo", "foo.bar"), nil, t)
+	checkError(ValidateMapping("*", "{{wildcard(1)}}"), nil, t)
+	checkError(ValidateMapping("foo.>", ">"), nil, t)
+	checkError(ValidateMapping("foo.*", "bar.{{wildcard(1)}}"), nil, t)
+	checkError(ValidateMapping("foo.>", "bar.>"), nil, t)
+	checkError(ValidateMapping("foo.*.>", "bar.{{wildcard(1)}}.>"), nil, t)
+	checkError(ValidateMapping("foo.*.bar", "bar.{{wildcard(1)}}.foo"), nil, t)
+	checkError(ValidateMapping("foo.bar.>", "foo.bar.foo.>"), nil, t)
+	checkError(ValidateMapping("*", "foo.{{wildcard(1)}}"), nil, t)
+	checkError(ValidateMapping("*", "foo.{{ wildcard(1) }}"), nil, t)
+	checkError(ValidateMapping("*", "foo.{{wildcard( 1 )}}"), nil, t)
+	checkError(ValidateMapping("*", "foo.{{partition(2,1)}}"), nil, t)
+	checkError(ValidateMapping("*.*", "foo.{{SplitFromLeft(2,1)}}"), nil, t)
+	checkError(ValidateMapping("*.*", "foo.{{SplitFromRight(2,1)}}"), nil, t)
+	checkError(ValidateMapping("*", "foo.{{unknown(1)}}"), ErrInvalidMappingDestination, t)
+	checkError(ValidateMapping("foo", "foo..}"), ErrInvalidMappingDestination, t)
+	checkError(ValidateMapping("foo", "foo. bar}"), ErrInvalidMappingDestinationSubject, t)
 }
 
 func TestSubjectToken(t *testing.T) {
@@ -1773,6 +1772,14 @@ func TestSublistHasInterest(t *testing.T) {
 	sl.Remove(qsub)
 }
 
+func TestSublistHasInterestOverlapping(t *testing.T) {
+	sl := NewSublistWithCache()
+	require_NoError(t, sl.Insert(newSub("stream.A.child")))
+	require_NoError(t, sl.Insert(newSub("stream.*")))
+	require_True(t, sl.HasInterest("stream.A.child"))
+	require_True(t, sl.HasInterest("stream.A"))
+}
+
 func TestSublistNumInterest(t *testing.T) {
 	sl := NewSublistWithCache()
 	fooSub := newSub("foo")
@@ -1983,128 +1990,6 @@ func TestSublistNumInterest(t *testing.T) {
 	sl.Remove(qsub)
 }
 
-func TestSublistInterestBasedIntersection(t *testing.T) {
-	st := stree.NewSubjectTree[struct{}]()
-	st.Insert([]byte("one.two.three.four"), struct{}{})
-	st.Insert([]byte("one.two.three.five"), struct{}{})
-	st.Insert([]byte("one.two.six"), struct{}{})
-	st.Insert([]byte("one.two.seven"), struct{}{})
-	st.Insert([]byte("eight.nine"), struct{}{})
-
-	require_NoDuplicates := func(t *testing.T, got map[string]int) {
-		for _, c := range got {
-			require_Equal(t, c, 1)
-		}
-	}
-
-	t.Run("Literals", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("one.two.six"))
-		sl.Insert(newSub("eight.nine"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 2)
-		require_NoDuplicates(t, got)
-	})
-
-	t.Run("PWC", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("one.two.*.*"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 2)
-		require_NoDuplicates(t, got)
-	})
-
-	t.Run("PWCOverlapping", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("one.two.*.four"))
-		sl.Insert(newSub("one.two.*.*"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 2)
-		require_NoDuplicates(t, got)
-	})
-
-	t.Run("PWCAll", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("*.*"))
-		sl.Insert(newSub("*.*.*"))
-		sl.Insert(newSub("*.*.*.*"))
-		require_True(t, sl.HasInterest("foo.bar"))
-		require_True(t, sl.HasInterest("foo.bar.baz"))
-		require_True(t, sl.HasInterest("foo.bar.baz.qux"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 5)
-		require_NoDuplicates(t, got)
-	})
-
-	t.Run("FWC", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("one.>"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 4)
-		require_NoDuplicates(t, got)
-	})
-
-	t.Run("FWCOverlapping", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("one.two.three.four"))
-		sl.Insert(newSub("one.>"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 4)
-		require_NoDuplicates(t, got)
-	})
-
-	t.Run("FWCAll", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub(">"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 5)
-		require_NoDuplicates(t, got)
-	})
-
-	t.Run("NoMatch", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("one"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 0)
-	})
-
-	t.Run("NoMatches", func(t *testing.T) {
-		got := map[string]int{}
-		sl := NewSublistNoCache()
-		sl.Insert(newSub("one"))
-		sl.Insert(newSub("eight"))
-		sl.Insert(newSub("ten"))
-		IntersectStree(st, sl, func(subj []byte, entry *struct{}) {
-			got[string(subj)]++
-		})
-		require_Len(t, len(got), 0)
-	})
-}
-
 func subsInit(pre string, toks []string) {
 	var sub string
 	for _, t := range toks {
